from pandas import DataFrame

from ..types import OpenBESSpecification, ENERGY_SOURCES


def month_for_day(day_number_in_year: int) -> int:
    """Calculate the month for a given day number in the year.
    Args:
        day_number_in_year (int): The day number in the year (1-365).
    Returns:
        int: The corresponding month (1-12).
    """
    if day_number_in_year <= 31:
        return 1
    elif day_number_in_year <= 59:
        return 2
    elif day_number_in_year <= 90:
        return 3
    elif day_number_in_year <= 120:
        return 4
    elif day_number_in_year <= 151:
        return 5
    elif day_number_in_year <= 181:
        return 6
    elif day_number_in_year <= 212:
        return 7
    elif day_number_in_year <= 243:
        return 8
    elif day_number_in_year <= 273:
        return 9
    elif day_number_in_year <= 304:
        return 10
    elif day_number_in_year <= 334:
        return 11
    else:
        return 12


# Blank DataFrame of each hour with month info, indexed by day of the year
HOURS_DF = DataFrame(
    [
        {"month": month_for_day(d), "day": d, "hour": h, "is_daytime": 8 <= h <= 22}
        for d in range(1, 366)
        for h in range(1, 25)
    ]
).set_index(["month", "day", "hour"])


class SimulationError(RuntimeError):
    """Base exception for simulation-level failures."""


class SimulationRequiredInputsError(SimulationError, ValueError):
    """Raised when required specification inputs are missing for a simulation."""


class EnergyUseSimulationInitError(SimulationError, ValueError):
    pass


def _resolve_required_input(spec: OpenBESSpecification, key: str):
    value = spec
    for part in key.split("."):
        value = getattr(value, part)
    return value


def missing_required_inputs(
    spec: OpenBESSpecification,
    required_inputs: list[str] | tuple[str, ...],
    getter=None,
) -> list[str]:
    if not required_inputs:
        return []
    resolver = getter or (lambda k: _resolve_required_input(spec, k))
    missing = []
    for key in required_inputs:
        try:
            value = resolver(key)
        except (AttributeError, KeyError, TypeError):
            missing.append(key)
            continue
        if value is None or value == "":
            missing.append(key)
    return missing


class HourlySimulation:
    """
    Base class for hourly simulations.

    Each instance is initialized with an OpenBESSpecification and contains an `_hours` property
    that holds a DataFrame representing each hour of the year.

    These simulations usually make heavy use of the `@property` decorator to define various calculated properties
    based on the specification and the hourly data.
    Properties will typically add columns to the `_hours` DataFrame as needed for various calculations,
    and usually return that DataFrame or specific columns from it.
    """

    spec: OpenBESSpecification
    _hours: DataFrame
    _required_inputs: tuple[str, ...] = ()
    _required_inputs_error_cls = SimulationRequiredInputsError

    def __init__(self, spec: OpenBESSpecification):
        self.spec = spec
        self.validate_required_inputs()
        self._hours = HOURS_DF.copy()

    def validate_required_inputs(
        self,
        required_inputs: list[str] | tuple[str, ...] | None = None,
        *,
        getter=None,
        context: str | None = None,
        error_cls=None,
    ) -> None:
        required = self._required_inputs if required_inputs is None else required_inputs
        missing = missing_required_inputs(self.spec, required, getter=getter)
        if not missing:
            return
        label = context or self.__class__.__name__
        exc_cls = error_cls or self._required_inputs_error_cls
        raise exc_cls(f"{label} missing required inputs: {', '.join(missing)}")


class EnergyUseSimulation(HourlySimulation):
    """
    Base class for hourly energy use simulations.

    Child classes should override the energy_use property to return a DataFrame representing
    the energy use in kW for each hour of the year, with a column for each of the ENERGY_SOURCES.

    The _energy_use property is initialized as a DataFrame of NaNs with the same index as _hours
    and columns for each ENERGY_SOURCES. self.energy_use should populate this DataFrame appropriately.
    """
    _required_inputs_error_cls = EnergyUseSimulationInitError

    def __init__(self, spec: OpenBESSpecification):
        super().__init__(spec)
        self._energy_use = DataFrame(
            index=self._hours.index, columns=list(ENERGY_SOURCES)
        ).astype(float)

    def get_param_or_spec(self, key: str):
        try:
            return getattr(self.spec.parameters, key)
        except AttributeError:
            return getattr(self.spec, key)

    @property
    def energy_use(self) -> DataFrame:
        """DataFrame of energy use in kW for each hour of the year, with a column for each ENERGY_SOURCES."""
        raise NotImplementedError(
            "Child classes must implement the energy_use property."
        )

    @property
    def annual_energy_use(self) -> DataFrame:
        return self.energy_use.sum().sum()

    @property
    def report(self) -> dict:
        raise NotImplementedError("Child classes must implement the report property.")



import logging
import time
from contextlib import contextmanager
from typing import Optional

_STORE: Optional[list[str]] = None
_PREFIX_STACK: list[str] = []
_T0: Optional[float] = None
_CONFIGURED = False
_RUN_ID = -1


def _prefix() -> str:
    return "".join(_PREFIX_STACK)


class ListHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        global _STORE
        if _STORE is None:
            return  # logging enabled but no store bound

        _STORE.append(self.format(record))


class RuntimeFormatter(logging.Formatter):
    def __init__(self, start_time: float = None, prefix: str = ""):
        super().__init__("%(runtime)s%(prefix)s%(levelname)s %(filename)s %(message)s")
        self.start_time = start_time if start_time is not None else time.perf_counter()

    def format(self, record: logging.LogRecord) -> str:
        # runtime
        if _T0 is None:
            record.runtime = ""
        else:
            record.runtime = f"{time.perf_counter() - _T0:8.3f}s "
        record.prefix = _prefix()
        if record.prefix:
            record.prefix = f"{record.prefix} "
        return super().format(record)


_HANDLER = ListHandler()
_HANDLER.setFormatter(RuntimeFormatter())


def configure_once(level=logging.INFO) -> None:
    global _CONFIGURED
    if _CONFIGURED:
        return
    logger = logging.getLogger("openbes")
    logger.setLevel(level)
    logger.addHandler(_HANDLER)
    _CONFIGURED = True


def bind(log_store: list[str], base_prefix: str = "") -> None:
    """
    Overwrites the global store/prefix/runtime anchor.
    Intended to be called by the root simulation during __init__.
    """
    global _STORE, _T0, _RUN_ID
    configure_once()

    _RUN_ID += 1
    run_id = _RUN_ID

    _STORE = log_store
    _T0 = time.perf_counter()

    _PREFIX_STACK.clear()
    if run_id > 0:
        _PREFIX_STACK.append(f"[run={run_id:03d}] ")
    if base_prefix:
        _PREFIX_STACK.append(base_prefix)


@contextmanager
def LogPrefix(prefix: str):
    _PREFIX_STACK.append(prefix or "")
    try:
        yield
    finally:
        _PREFIX_STACK.pop()


def getLogger(name: str) -> logging.Logger:
    configure_once()
    return logging.getLogger(name)



# generated by datamodel-codegen:
#   filename:  OpenBES.schema.json
#   timestamp: 2026-03-02T10:49:57+00:00

from __future__ import annotations
from enum import StrEnum
from pydantic import BaseModel, ConfigDict, Field, RootModel


class ENERGYSOURCES(StrEnum):
    Electricity = "Electricity"
    Diesel = "Diesel"
    LPG = "LPG"
    Natural_gas = "Natural gas"
    Biomass = "Biomass"
    Pellets = "Pellets"


class COOLINGSYSTEMTYPES(StrEnum):
    Heat_pump = "Heat pump"


class HEATINGSYSTEMTYPES(StrEnum):
    Electric_heating__radiators_ = "Electric heating (radiators)"
    Electric_boiler = "Electric boiler"
    Conventional_boiler = "Conventional boiler"
    Low_temperature_boiler = "Low temperature boiler"
    Condensing_boiler = "Condensing boiler"
    Biomass_boiler = "Biomass boiler"
    Heat_pump = "Heat pump"


class METEOROLOGICALFILE(StrEnum):
    openbes___USA_Denver_725650TYCST_epw = "openbes://USA_Denver_725650TYCST.epw"
    openbes___UK_Oxford_GBR_ENG_RAF_Benson_036580_TMYx_2007_2021_epw = (
        "openbes://UK_Oxford_GBR_ENG_RAF.Benson.036580_TMYx.2007-2021.epw"
    )
    openbes___SPAIN_Sevilla_083910_SWEC_epw = "openbes://SPAIN_Sevilla.083910_SWEC.epw"
    openbes___SPAIN_Madrid_082210_SWEC_epw = "openbes://SPAIN_Madrid.082210_SWEC.epw"


class TERRAINCLASS(StrEnum):
    Open_terrain = "Open terrain"
    Country = "Country"
    Urban = "Urban"


class SUPPORTEDCOUNTRIES(StrEnum):
    Spain = "Spain"
    USA = "USA"
    France = "France"
    UK = "UK"
    Other = "Other"


class WindowCounts(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    front: list[int] | None = Field(
        default=None,
        description="Array of window counts per floor on the front side of the building.",
    )
    back: list[int] | None = Field(
        default=None,
        description="Array of window counts per floor on the back side of the building.",
    )
    left: list[int] | None = Field(
        default=None,
        description="Array of window counts per floor on the left side of the building.",
    )
    right: list[int] | None = Field(
        default=None,
        description="Array of window counts per floor on the right side of the building.",
    )


class Building(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    area: str | None = Field(
        default=None, description="Street/region the building is in."
    )
    height: float | None = Field(
        default=None, description="Height of the building in m."
    )
    length: float = Field(..., description="Length of the building in m.")
    width: float = Field(..., description="Width of the building in m.")
    name: str | None = Field(default=None, description="Name of the building")
    type: str | None = Field(
        default=None, description="Type of building (e.g. Office, School)."
    )
    floor_count: int | None = Field(
        default=1, description="Number of floors in the building."
    )
    floor_to_ceiling_height: float | None = Field(
        default=2.7, description="Floor to ceiling height in m."
    )
    slab_thickness: float | None = Field(
        default=0.2, description="Slab thickness in m."
    )
    orientation_angle: float | None = Field(
        default=0.0,
        description="Orientation angle of the building in degrees clockwise from North=0.",
    )
    roof_angle: float | None = Field(
        default=0.0, description="Roof angle in degrees from horizontal."
    )
    solar_external_shading_summer: float | None = Field(
        default=0.0,
        description="External solar shading factor for summer (Summer = air temp > 22C).",
    )
    solar_external_shading_winter: float | None = Field(
        default=0.0,
        description="External solar shading factor for winter (Winter = air temp <= 22C).",
    )
    terrain_class: TERRAINCLASS | None = Field(
        default=None, description="Terrain class around the building."
    )
    thermal_bridge_facade_ground: bool | None = Field(
        default=None,
        description="Whether to model the thermal bridge factor for facade to ground junctions.",
    )
    thermal_bridge_facade_intermediate: bool | None = Field(
        default=None,
        description="Whether to model the thermal bridge factor for facade to intermediate floor junctions.",
    )
    thermal_bridge_facade_roof: bool | None = Field(
        default=None,
        description="Whether to model the thermal bridge factor for facade to roof junctions.",
    )
    thermal_bridge_shading: bool | None = Field(
        default=None,
        description="Whether to model the thermal bridge factor for shading devices.",
    )
    thermal_bridge_window: bool | None = Field(
        default=None,
        description="Whether to model the thermal bridge factor for windows.",
    )
    uvalue_facade: float | None = Field(
        default=0.0, description="U-value of the facade in W/m2K."
    )
    uvalue_roof: float | None = Field(
        default=0.0, description="U-value of the roof in W/m2K."
    )
    uvalue_window: float | None = Field(
        default=0.0, description="U-value of the windows in W/m2K."
    )
    uvalue_floor: float | None = Field(
        default=0.0, description="U-value of the floor in W/m2K."
    )
    window_frame_factor: float | None = Field(
        default=0.0, description="Window frame factor (fraction)."
    )
    window_gvalue: float | None = Field(
        default=0.0, description="Window g-value (fraction)."
    )
    window_height: float | None = Field(default=0.0, description="Window height in m.")
    window_length: float | None = Field(default=0.0, description="Window length in m.")
    window_counts: WindowCounts | None = Field(
        default=None,
        description="A map of building side to array of window counts per floor.",
    )


class Duration(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    start: int = Field(..., description="Start hour (1-24).")
    end: int = Field(..., description="End hour (1-24).")


class Zone(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    name: str = Field(..., description="Zone name")
    areas: list[float] = Field(..., description="Zone area for each floor in m2")
    conditioned: bool | None = Field(
        default=True, description="Whether this zone uses HVAC."
    )
    active_hours: Duration | None = Field(
        default=None, description="Times the zone is occupied each day."
    )


class ZoneSimultaneity(RootModel[dict[str, float]]):
    root: dict[str, float]


class OccupationSchedule(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    January: bool | None = False
    February: bool | None = False
    March: bool | None = False
    April: bool | None = False
    May: bool | None = False
    June: bool | None = False
    July: bool | None = False
    August: bool | None = False
    September: bool | None = False
    October: bool | None = False
    November: bool | None = False
    December: bool | None = False
    Monday: bool | None = False
    Tuesday: bool | None = False
    Wednesday: bool | None = False
    Thursday: bool | None = False
    Friday: bool | None = False
    Saturday: bool | None = False
    Sunday: bool | None = False


class Range(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    min: float = Field(..., description="Minimum value of the range.")
    max: float = Field(..., description="Maximum value of the range.")


class Consumption(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    January: float | None = 0.0
    February: float | None = 0.0
    March: float | None = 0.0
    April: float | None = 0.0
    May: float | None = 0.0
    June: float | None = 0.0
    July: float | None = 0.0
    August: float | None = 0.0
    September: float | None = 0.0
    October: float | None = 0.0
    November: float | None = 0.0
    December: float | None = 0.0


class HeatingSystem(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    type: HEATINGSYSTEMTYPES | None = None
    energy_source: ENERGYSOURCES = Field(
        ..., description="Energy source for this heating system."
    )
    efficiency_cop: float = Field(..., description="COP or efficiency (if applicable).")
    min_demand: float | None = Field(
        default=0.0, description="Minimum demand threshold."
    )
    nominal_capacity: float = Field(..., description="Nominal capacity.")
    count: int | None = Field(default=1, description="Number of identical units.")
    active_hours: Duration = Field(..., description="Times the system is on each day.")
    simultaneity: ZoneSimultaneity | None = None


class CoolingSystem(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    type: COOLINGSYSTEMTYPES | None = None
    energy_source: ENERGYSOURCES
    efficiency_ratio: float = Field(
        ..., description="EER/COP or energy efficiency ratio."
    )
    min_demand: float | None = Field(default=None, description="Minimum demand.")
    nominal_capacity: float = Field(..., description="Nominal capacity.")
    sensible_nominal_capacity: float = Field(..., description="Sensible capacity.")
    count: int | None = Field(default=1, description="Number of identical units.")
    active_hours: Duration = Field(..., description="Times the system is on each day.")
    simultaneity: ZoneSimultaneity | None = None


class VentilationSystem(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    type: float | None = Field(default=None, description="Optional numeric type code.")
    energy_source: ENERGYSOURCES
    airflow: float = Field(..., description="Airflow (m3/s).")
    heat_recovery_efficiency: float | None = Field(
        default=None, description="Heat recovery efficiency."
    )
    rated_input_power: float | None = Field(
        default=None, description="Rated input power."
    )
    ventilated_area: float | None = Field(
        default=1.0, description="Proportion of area ventilated."
    )
    active_hours: Duration | None = Field(
        default_factory=lambda: Duration.model_validate({"start": 1, "end": 24}),
        description="Times the system is on each day.",
    )


class Ballast(StrEnum):
    Electronic_ballast = "Electronic ballast"
    Ferromagnetic_ballast = "Ferromagnetic ballast"


class LightingSystemTech(StrEnum):
    Tubular_fluorescent_T8 = "Tubular fluorescent T8"
    Tubular_fluorescent_T5 = "Tubular fluorescent T5"
    Compact_fluorescent = "Compact fluorescent"
    Incandescent = "Incandescent"
    Halogen = "Halogen"
    Mercury_vapor = "Mercury vapor"
    Sodium_vapour = "Sodium vapour"
    Metal_halide = "Metal halide"
    Induction = "Induction"
    LED = "LED"


class HotWaterSystem(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    demand: Consumption = Field(..., description="Hot water demand (litres).")
    reference_temperature: float = Field(..., description="Reference temperature (°C).")
    supply_temperature: float = Field(..., description="Supply temperature (°C).")
    energy_source: ENERGYSOURCES = Field(
        ..., description="Energy source for this hot water system."
    )
    efficiency_cop: float = Field(
        ..., description="Efficiency of the hot water system."
    )
    nominal_capacity: float = Field(
        ..., description="Nominal capacity of the hot water system."
    )
    type: float | None = Field(default=None, description="[UNUSED]")


class Courtyard(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    length: float = Field(..., description="Courtyard length (m).")
    width: float = Field(..., description="Courtyard width (m).")
    count: int | None = Field(
        default=1, description="How many courtyards of this geometry exist."
    )


class OpenCourtyard(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    depth: float = Field(..., description="Courtyard depth (m).")
    count: int | None = Field(
        default=1, description="How many courtyards of this geometry exist."
    )


class OpenCourtyardMap(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    front: OpenCourtyard | None = None
    back: OpenCourtyard | None = None
    left: OpenCourtyard | None = None
    right: OpenCourtyard | None = None


class HeatCapacity(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    Am: float = Field(..., description="Mass factor??????")
    Cm: float = Field(..., description="Capacitance??????")


class OpenBESParameters(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    altitude: float | None = Field(
        default=None,
        description="Altitude (m). [UNUSED: the EPW file's 'altitude' is used instead]",
    )
    include_appliances: bool | None = Field(
        default=True, description="Model appliance consumption."
    )
    cooling_load_factor: float | None = Field(
        default=1.0, description="Multiplier applied to cooling loads."
    )
    density_of_air: float | None = Field(
        default=1.211, description="Air density (kg/m3)."
    )
    facade_absorption_coefficient: float | None = Field(
        default=0.6, description="Facade solar absorption coefficient."
    )
    facade_correction_factor: float | None = Field(
        default=1.0, description="Facade correction factor."
    )
    facade_emissivity: float | None = Field(
        default=0.9, description="Facade emissivity value."
    )
    floor_correction_factor: float | None = Field(
        default=1.0, description="Floor correction factor."
    )
    heat_capacity_correction_factor: float | None = Field(
        default=1.0, description="Heat capacity correction factor."
    )
    heat_capacity_joule: float | None = Field(
        default=None, description="Heat capacity of air in J/kgK."
    )
    heating_load_factor: float | None = Field(
        default=1.0, description="Multiplier applied to heating loads."
    )
    infiltration_correction_factor: float | None = Field(
        default=1.0, description="Correction factor for infiltration."
    )
    leakage_air_flow_dependent: float | None = Field(
        default=0.0001,
        description="Leakage air flow rate dependent on other factors (m3/s).",
    )
    include_lighting: bool | None = Field(
        default=True, description="Whether lighting is modelled."
    )
    nia_gba_ratio: float | None = Field(
        default=0.95, description="Net internal area / gross building area fraction."
    )
    include_occupancy: bool | None = Field(
        default=True, description="Whether occupancy schedules are modelled."
    )
    output_csv_precision: int | None = Field(
        default=2, description="Number of decimal places in CSV output."
    )
    pressure_of_air: float | None = Field(
        default=None, description="Air pressure. [UNUSED]"
    )
    roof_absorption_coefficient: float | None = Field(
        default=0.8, description="Roof solar absorption coefficient."
    )
    roof_correction_factor: float | None = Field(
        default=1.0, description="Roof correction factor."
    )
    roof_emissivity: float | None = Field(default=0.9, description="Roof emissivity.")
    shading_correction_factor: float | None = Field(
        default=1.0, description="Shading correction factor."
    )
    specific_heat_of_air: float | None = Field(
        default=1.015, description="Specific heat of air."
    )
    view_factor_to_sky_facade: float | None = Field(
        default=0.5, description="View factor to sky from facade."
    )
    view_factor_to_sky_roof: float | None = Field(
        default=1.0, description="View factor to sky from roof."
    )
    window_correction_factor: float | None = Field(
        default=1.0, description="Window correction factor."
    )
    window_angular_correction_factors: list[float] | None = Field(
        default=[1.0, -0.189, 0.644, -0.596, 0.0],
        description="Angular correction factors for windows.",
        max_length=5,
        min_length=5,
    )


class LightingControl(StrEnum):
    Manual = "Manual"
    Automatic = "Automatic"


class SpaceThermalDemandResult(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    demand_total: float | None = Field(
        default=None, description="Annual energy required for heating/cooling in kWh."
    )
    demand_scaled: float | None = Field(
        default=None,
        description="Annual energy required for heating/cooling in kWh/m2.",
    )
    demand_on_all_year: float | None = Field(
        default=None,
        description="Annual energy demanded in kWh/m2 if the system is on all year round.",
    )
    load_csv: str | None = Field(
        default=None, description="Load in kWh/m2 by calendar month."
    )
    load_duraction_csv: str | None = Field(
        default=None, description="Load in kW by annual percentile."
    )


class VentilationSystemResult(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    energy_demand: float | None = Field(
        default=None, description="Energy demanded in kWh/m2"
    )
    peak_load: float | None = Field(
        default=None, description="Maximum hourly energy required in kW."
    )
    sfp: float | None = Field(
        default=None, description="Specific fan power in kW/m3/s."
    )
    mechanical_ventilation_rate: float | None = Field(
        default=None,
        description="Mechanical ventilation rate in m3/h per m2. Recommended: 3 m3/h m2.",
    )
    ventilation_rate: float | None = Field(
        default=None,
        description="Ventilation rate in l/s m2. Reference: 0.83l/s m2 or 12.5l/s per person.",
    )
    ach: float | None = Field(default=None, description="Air changes per hour.")


class ModelValidation(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    energy_use_csv: str | None = Field(
        default=None,
        description="Energy used for each calendar month in kWh. Compare to user-supplied values in `OpenBESSpecification.*_consumption`.",
    )
    nmbe: float | None = Field(default=None, description="Normalised mean bias error.")
    cv_rmse: float | None = Field(
        default=None,
        description="Coefficient of variation of the root mean square deviation.",
    )
    r2: float | None = Field(default=None, description="R-squared model fit.")


class Month(StrEnum):
    January = "January"
    February = "February"
    March = "March"
    April = "April"
    May = "May"
    June = "June"
    July = "July"
    August = "August"
    September = "September"
    October = "October"
    November = "November"
    December = "December"


class HourPeak(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    value: float | None = Field(default=None, description="The statistic in question.")
    month: Month | None = None
    day: float | None = Field(default=None, description="Day of the month.")
    hour: float | None = Field(default=None, description="Hour of the day.")


class LocationSimulationOutput(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    altitude: float | None = Field(
        default=None, description="Altitude as read from the weather file."
    )
    solstice_ghr_csv: str | None = Field(
        default=None,
        description="Global Horizontal Radiation (Wh/m2) for the longest and shortest days of the year (21 Dec, 21 Jun).",
    )
    max_outdoor_temperature: HourPeak | None = None
    min_outdoor_temperature: HourPeak | None = None
    climate_quantiles_csv: str | None = Field(
        default=None,
        description="Outdoor air temperature percentiles from climate data.",
    )
    degree_days_csv: str | None = Field(
        default=None, description="Heating/cooling degree days by month."
    )
    annual_incident_solar_radiation_csv: str | None = Field(
        default=None,
        description="Annual incident solar radiation in kWh/m2 for each 1/8th compass point.",
    )


class GeometrySimulationOutput(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    gross_building_area: float | None = Field(
        default=None, description="Total building area in m2."
    )
    conditioned_floor_area: float | None = Field(
        default=None,
        description="Floor area subject to HVAC systems. Used as the divisor in most /m2 statistics.",
    )
    indoor_air_volume: float | None = Field(
        default=None,
        description="Volume of air in the building's conditioned areas in m3.",
    )
    indoor_air_heat_capacity: float | None = Field(
        default=None, description="Indoor air heat capacity in kJ/K"
    )
    building_geometry_csv: str | None = Field(
        default=None,
        description="Building area by floor and category in m2, and window-to-wall ratio (unitless).",
    )
    building_geometry_orientation_csv: str | None = Field(
        default=None,
        description="Building facade, window, and roof area in m2 by category and orientation.",
    )


class ClimateSimulationOutput(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    external_internal_temperature_csv: str | None = Field(
        default=None,
        description="Internal (HVAC) and external temperatures for each hour of the year.",
    )
    max_indoor_temperature: HourPeak | None = None
    min_indoor_temperature: HourPeak | None = None
    mean_indoor_temperature: float | None = Field(
        default=None, description="Mean indoor temperature over the year."
    )
    discomfort_hours_percent: float | None = Field(
        default=None,
        description="The number of hours when the building is occupied and it would be outside 18-26C if HVAC were not operating.",
    )
    discomfort_hours_percent_summer: float | None = Field(
        default=None,
        description="The number of hours when the building is occupied and it would be outside 18-26C if HVAC were not operating during June, July, and August.",
    )
    heat_exchange_breakdown_csv: str | None = Field(
        default=None,
        description="The heat loss/gain in kWh/m2 for each month, broken down by transmission category.",
    )
    space_thermal_demand_csv: str | None = Field(
        default=None,
        description="The heating and cooling energy requirement in kWh/m2 for each month of the year based on heat, mass balance, and occupancy.",
    )
    infiltration_ach: float | None = Field(
        default=None, description="Air changes/hour from air infiltration."
    )
    natural_ach: float | None = Field(
        default=None, description="Air changes/hour from natural ventilation."
    )
    heating_demand: SpaceThermalDemandResult | None = None
    cooling_demand: SpaceThermalDemandResult | None = None
    running_average_outside_temp_csv: str | None = Field(
        default=None,
        description="The 7-day weighted running mean of the outside temperature for each hour of the year. The first 6 days copy the value from day 7.",
    )
    overheating_limits_csv: str | None = Field(
        default=None,
        description="Category I, II, and III overheating limits in °C at different outdoor temperatures.",
    )
    solar_heat_gains_csv: str | None = Field(
        default=None,
        description="Solar heat gains in kWh by orientation and category. For kWh/m2 divide by relevant `building_facade_csv` area.",
    )
    window_transmissivity_coefficient_csv: str | None = Field(
        default=None,
        description="Annual transmissivity coefficient of windows by orientation.",
    )


class VentilationSimulationOutput(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    ventilation_systems: list[VentilationSystemResult] | None = None


class LightingSimulationOutput(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    lighting_demand: float | None = Field(
        default=None, description="Energy demand for lighting in kWh/m2."
    )
    lighting_peak_load: float | None = Field(
        default=None, description="Peak lighting load in kW."
    )
    lighting_load_ratio: float | None = Field(
        default=None, description="Lighting energy required in W/m2."
    )


class HotWaterSimulationOutput(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    hot_water_demand: float | None = Field(
        default=None, description="Energy demand for water heating in kWh/m2."
    )


class BuildingEnergySimulationOutput(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    other_energy_use_electricity: float | None = Field(
        default=None, description="Miscellaneous electricity use in kWh/m2."
    )
    other_energy_use_gas: float | None = Field(
        default=None, description="Miscellaneous natural gas usage in kWh/m2."
    )
    on_site_electricity_generated: float | None = Field(
        default=None, description="Renewable electricity generated on site in kWh/m2."
    )
    on_site_electricity_used: float | None = Field(
        default=None,
        description="Renewable electricity generated and consumed on site in kWh/m2.",
    )
    on_site_electricity_fraction: float | None = Field(
        default=None,
        description="Renewable energy consumption / Final energy consumption",
    )
    all_renewable_fraction: float | None = Field(
        default=None,
        description="Fraction of final energy consumption drawn from renewable energy sources including both on- and off-site renewable energy sources.",
    )
    final_energy_consumption: float | None = Field(
        default=None, description="Final energy consumption in kWh/m2."
    )
    primary_energy_consumption: float | None = Field(
        default=None, description="Primary energy consumption in kWh/m2."
    )
    non_renewable_primary_energy_consumption: float | None = Field(
        default=None, description="Non-renewable primary energy consumption in kWh/m2."
    )
    co2_equivalent_emissions: float | None = Field(
        default=None, description="CO2 equivalent emissions in kg CO2 eq/m2."
    )
    final_energy_consumption_csv: str | None = Field(
        default=None,
        description="Final energy consumption in kWh for each energy use category. Divide by `conditioned_floor_area` for kWh/m2.",
    )
    electricity_validation: ModelValidation | None = None
    gas_validation: ModelValidation | None = None


class OpenBESMetaData(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    version: str = Field(..., description="Version of Open BES used.")
    timestamp: str | None = Field(
        default=None, description="Timestamp of when the simulation was run."
    )
    EPW_file_checksum: str | None = Field(
        default=None, description="MD5 Checksum of the EPW file used in the simulation."
    )


class LightingSystem(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    tech: LightingSystemTech
    ballast: Ballast | None = Field(default=None, description="Ballast type.")
    lamp_number: int = Field(..., description="Number of lamps.")
    lamp_power: float = Field(..., description="Lamp power (W).")
    luminary_number: int = Field(..., description="Number of luminaires.")
    name: str | None = Field(
        default=None, description="Optional human label for the system."
    )
    active_hours: Duration = Field(
        ..., description="Times the lighting is on each day."
    )
    count: int | None = Field(
        default=1, description="Count of similar lighting systems aggregated together."
    )
    simultaneity_factor: float = Field(
        ..., description="Simultaneity factor (fraction)."
    )


class OpenBESSpecification(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    parameters: OpenBESParameters | None = Field(
        default=None,
        description="Collapsed parameters object. Use this to hold heating/cooling/ventilation/lighting/courtyards arrays.",
    )
    building: Building | None = Field(
        default=None, description="Building general information."
    )
    zones: list[Zone] | None = Field(default=None, description="Zones in the building.")
    occupation_schedule: OccupationSchedule | None = Field(
        default=None, description="Occupation schedule for the building."
    )
    heating_systems: list[HeatingSystem] | None = Field(
        default=None, description="Heating systems in the building."
    )
    cooling_systems: list[CoolingSystem] | None = Field(
        default=None, description="Cooling systems in the building."
    )
    ventilation_systems: list[VentilationSystem] | None = Field(
        default=None, description="Ventilation systems in the building."
    )
    lighting_systems: list[LightingSystem] | None = Field(
        default=None, description="Lighting systems in the building."
    )
    hot_water_systems: list[HotWaterSystem] | None = Field(
        default=None, description="Hot water systems in the building."
    )
    courtyards: list[Courtyard] | None = Field(
        default=None, description="Courtyards in the building."
    )
    open_courtyards: OpenCourtyardMap | None = Field(
        default=None, description="Open courtyards on each side of the building."
    )
    appliances_load: float | None = Field(
        default=0.0, description="Load from appliances (Wh/m2)."
    )
    biomass_consumption: Consumption | None = Field(
        default=None, description="Amount of biomass consumed by the building. [UNUSED]"
    )
    biomass_pellets_consumption: Consumption | None = Field(
        default=None,
        description="Amount of biomass pellets consumed by the building. [UNUSED]",
    )
    building_standby_electricity_consumption: Consumption | None = Field(
        default=None,
        description="Electrical standby load for the building.",
    )
    diesel_consumption: Consumption | None = Field(
        default=None, description="Amount of diesel consumed by the building. [UNUSED]"
    )
    electricity_consumption: Consumption | None = Field(
        default=None,
        description="Amount of electricity consumed by the building. [UNUSED]",
    )
    other_electricity_consumption: Consumption | None = Field(
        default=None,
        description="Amount of other electricity consumed by the building. [UNUSED]",
    )
    energy_generated: Consumption | None = Field(
        default=None, description="Amount of energy generated by the building. [UNUSED]"
    )
    energy_used: Consumption | None = Field(
        default=None, description="Amount of energy used by the building. [UNUSED]"
    )
    other_gas_consumption: Consumption | None = Field(
        default=None,
        description="Amount of other gas consumed by the building. [UNUSED]",
    )
    LPG_consumption: Consumption | None = Field(
        default=None, description="Amount of LPG consumed by the building. [UNUSED]"
    )
    natural_gas_consumption: Consumption | None = Field(
        default=None,
        description="Amount of natural gas consumed by the building. [UNUSED]",
    )
    country: SUPPORTEDCOUNTRIES | None = None
    heat_capacity: HeatCapacity | None = Field(
        default=None, description="Heat capacity properties of the building."
    )
    holiday: bool | None = Field(
        default=None, description="Whether the building is open on public holidays."
    )
    leakage_air_flow: float | None = Field(
        default=None, description="Leakage air flow rate (m3/s)."
    )
    leakage_air_flow_independent: float | None = Field(
        default=0.0,
        description="Leakage air flow rate independent of other factors (m3/s).",
    )
    lighting_control: LightingControl | None = Field(
        default="Automatic", description="Type of lighting control system."
    )
    lighting_active_hours: Duration | None = Field(
        default=None, description="Daily active hours for lighting."
    )
    lighting_simultaneity_factor: float | None = Field(
        default=0.2, description="Simultaneity factor for lighting."
    )
    location: str | None = Field(
        default=None,
        description="Location identifier or name (used to map to meteorological file).",
    )
    max_building_occupation: int | None = Field(
        default=None,
        description="Maximum occupancy of the building in number of people.",
    )
    meteorological_file_path: METEOROLOGICALFILE | None = Field(
        default=None, description="Meteorological filename or identifier."
    )
    natural_ventilation_night: float | None = Field(
        default=None, description="Night-time natural ventilation rate."
    )
    setpoint_temperature_day: Range | None = Field(
        default_factory=lambda: Range.model_validate({"min": 20.0, "max": 20.0}),
        description="Daytime temperature setpoint range.",
    )
    setpoint_temperature_night: Range | None = Field(
        default_factory=lambda: Range.model_validate({"min": 20.0, "max": 20.0}),
        description="Night-time temperature setpoint range.",
    )
    typical_occupation: int | None = Field(
        default=None,
        description="Typical occupancy of the building in number of people.",
    )


class ThermalSystemResult(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    conditioned_area: float | None = Field(
        default=None,
        description="Conditioned equivalent area in m2 (conditioned area adjusted for simultaneity).",
    )
    energy_demand: float | None = Field(
        default=None,
        description="Energy demanded in kWh/m2 when specific zonal occupancy is simulated.",
    )
    system_usage: float | None = Field(
        default=None,
        description="Energy demand divided by conditioned area (no simultaneity adjustment) in kWh/m2",
    )
    peak_load: HourPeak | None = None
    peak_capacity: float | None = Field(
        default=None, description="Maximum hourly energy consumed in kW."
    )
    peak_ratio: float | None = Field(
        default=None,
        description="Sensible peak capacity for cooling systems. For heating systems, peak capacity/load at 99.6th percentile.",
    )


class HeatingSimulationOutput(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    heating_systems: list[ThermalSystemResult] | None = None


class CoolingSimulationOutput(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    cooling_systems: list[ThermalSystemResult] | None = None


class OpenBESOutput(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    location_simulation_output: LocationSimulationOutput | None = None
    geometry_simulation_output: GeometrySimulationOutput | None = None
    climate_simulation_output: ClimateSimulationOutput | None = None
    ventilation_simulation_output: VentilationSimulationOutput | None = None
    heating_simulation_output: HeatingSimulationOutput | None = None
    cooling_simulation_output: CoolingSimulationOutput | None = None
    lighting_simulation_output: LightingSimulationOutput | None = None
    hot_water_simulation_output: HotWaterSimulationOutput | None = None
    building_energy_simulation_output: BuildingEnergySimulationOutput | None = None


class OpenBESCase(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    inputs: OpenBESSpecification
    outputs: OpenBESOutput | None = None
    log: list[str] | None = Field(
        default=None, description="Log output from the simulation run."
    )
    meta: OpenBESMetaData | None = None



from .. import logging
from typing import List

import numpy as np
from pandas import Series

from .base import EnergyUseSimulation, EnergyUseSimulationInitError, SimulationError
from .climate import ClimateSimulation
from .geometry import BuildingGeometry
from .lighting import LightingSimulation
from .occupancy import OccupationSimulation
from .reporting import find_hour_peak, output_precision
from .ventilation import VentilationSimulation
from ..schemas import HeatingSimulationOutput, ThermalSystemResult
from ..types import (
    OpenBESSpecification,
    HEATING_SYSTEM_TYPES,
)

logger = logging.getLogger(__name__)


class HeatingSimulationError(SimulationError):
    """Raised when heating report generation fails."""

MIN_HEATING_CAPACITY = 0.01  # [Hardcoded in Inputs cell F287]
MAX_HEATING_CAPACITY_LIMIT = 1.5  # [Hardcoded in Hourly Simulation cell ES114]


class HeatingSystemSimulation(EnergyUseSimulation):
    """A class to simulate a heating system's energy consumption.

    A heating system's energy use is determined by its capacity, the heating load, and the efficiency of the system.
    The heating load is determined by the difference between the indoor temperature and the outdoor temperature,
    as well as the area being heated and the occupancy of the building.

    Note that the heating system load is dependent on the heating system and the area it has to heat,
    as opposed to the Climate Simulation's heating demand, which is a model of the total heating demand of the
    building based on the climate and building characteristics,
    independent of any heating system that happens to be installed.

    [Cell references are for System 1]
    """

    system_number: int
    _nominal_consumption: float = None
    _nominal_capacity: float = None
    _sensible_nominal_capacity: float = None
    _efficiency: float = None
    _area: float = None
    _Ts_int: float = None
    _Th_int: float = None

    def __init__(
        self,
        spec: OpenBESSpecification,
        system_number: int = 1,
        climate: ClimateSimulation = None,
    ):
        super().__init__(spec)
        self.system_number = system_number
        self.validate_required_inputs(
            required_inputs=["energy_source", "efficiency_cop", "nominal_capacity"],
            getter=self._attr,
            context=f"Heating system {system_number}",
            error_cls=EnergyUseSimulationInitError,
        )
        try:
            self.climate = climate or ClimateSimulation(spec)
            self.geometry = self.climate.geometry
            self.occupancy = self.climate.occupancy
        except SimulationError as exc:
            raise EnergyUseSimulationInitError(
                f"Failed to initialize HeatingSystemSimulation due to error in dependent simulation: {exc}"
            ) from exc

    def _attr(self, attr_name: str):
        return self.get_param_or_spec(f"heating_system{self.system_number}_{attr_name}")

    @property
    def area(self) -> float:
        """Total area affected by heating system in m2."""
        if self._area is None:
            areas = self.geometry.conditioned_floor_areas.groupby("zone").sum()
            simultaneity = [
                self._attr(f"simultaneity_factor_{z.value.split('_')[0]}")
                for z in areas.index
            ]
            self._area = (areas * simultaneity).sum()
        return self._area

    @property
    def ratio(self) -> "Series[float]":
        """Hourly proportion of total heating capacity that is enabled.

        The system will typically be 100% enabled during occupied hours.
        """
        return self.occupancy.is_occupied.astype(float)

    def _get_target_temp(self) -> "np.array[float]":
        """Get the target temperature for heating based on thresholds and air temperature."""
        summer_temp = np.array(self.climate.set_point_temperature["min_temp_set_point"])
        winter_temp = np.array(self.climate.set_point_temperature["max_temp_set_point"])
        return np.where(
            self.climate.air_free_temp < winter_temp,
            winter_temp,
            np.where(
                self.climate.air_free_temp > summer_temp,
                summer_temp,
                self.climate.air_free_temp,
            ),
        )

    @property
    def phi_hc_nd_actual(self) -> "Series[float]":
        """Heating load per unit area in W/m2.
        [Hourly simulation column EO]

        The system is on when
        - the system is enabled (ratio > 0.0)
        - temperature < target temperature

        When on, the system operates at the capacity required to bring the temperature within tolerance.
        """
        if "phi_hc_nd_actual" not in self._hours.columns:
            target_temp = self._get_target_temp()
            self._hours["phi_hc_nd_actual"] = (
                (
                    10
                    * (target_temp - self.climate.air_free_temp_hc_0)
                    / (
                        self.climate.air_free_temp_hc_10
                        - self.climate.air_free_temp_hc_0
                    )
                )
                * self.ratio
                * (target_temp > self.climate.air_free_temp)
            )
        return self._hours["phi_hc_nd_actual"]

    @property
    def phi_h_nd_ac(self) -> "Series[float]":
        """Heating load per unit area in W/m2.
        [Hourly simulation column EP]
        """
        if "phi_h_nd_ac" not in self._hours.columns:
            self._hours["phi_h_nd_ac"] = (
                self.phi_hc_nd_actual.apply(
                    lambda r: (
                        max(r, 0.0)
                        if r
                        > getattr(
                            self.spec.parameters,
                            f"heating_system{self.system_number}_min_demand",
                        )
                        else 0.0
                    )
                )
                * self.spec.parameters.heating_load_factor
            )
        return self._hours["phi_h_nd_ac"]

    @property
    def demand(self) -> "Series[float]":
        """Hourly heating demand in kW.
        [Hourly simulation column ES]
        """
        if "demand" not in self._hours.columns:
            self._hours["demand"] = (self.phi_h_nd_ac * self.area) / 1000  # W -> kW
            self._hours["demand"] = self._hours["demand"].clip(
                upper=MAX_HEATING_CAPACITY_LIMIT * self.capacity
            )
        return self._hours["demand"]

    @property
    def capacity(self):
        """Heating system capacity in kW.
        [Hourly simulation cell EY92]
        """
        return (
            self.nominal_capacity * 1
        )  # [1 is the capacity factor, hardcoded in EY93]

    @property
    def number(self) -> int:
        """Number of systems of this type installed."""
        return self._attr("number")

    @property
    def min_demand(self) -> float:
        """Minimum demand for the system in kW.
        [Hourly simulation cell EY91]
        """
        return self._attr("min_demand")

    @property
    def nominal_capacity(self) -> float:
        """ """
        if self._nominal_capacity is None:
            self._nominal_capacity = (
                max(self._attr("nominal_capacity"), MIN_HEATING_CAPACITY) * self.number
            )
        return self._nominal_capacity

    @property
    def ren_fcp_potential(self) -> "Series[float]":
        """Part-load efficiency factor for heating system. ??????
        [Hourly simulation column EY]
        """
        if "ren_fcp_potential" not in self._hours.columns:
            f1 = 1.0914  # [Hardcoded in Hourly simulation cell EY102]
            f2 = -0.0914  # [Hardcoded in Hourly simulation cell EY103]
            self._hours["ren_fcp_potential"] = (
                self.demand > 0.0
            ) * f1 + f2 * self.fcp_p
        return self._hours["ren_fcp_potential"]

    @property
    def Rend(self) -> "Series[float]":
        """Efficiency of the heating system.
        [Hourly simulation column EZ]
        """
        if "Rend" not in self._hours.columns:
            ren_T = 1.0  # [Hardcoded in Hourly simulation cell EY96]
            ren_fcp_t = 1.0  # [Hardcoded in Hourly simulation cell EY98]
            self._hours["Rend"] = (
                self._attr("efficiency_cop")
                * ren_T
                * self.ren_fcp_potential
                * ren_fcp_t
            )
        return self._hours["Rend"]

    @property
    def fcp_p(self) -> "Series[float]":
        """Part-load factor for heating system. ??????
        [Hourly simulation column EX]
        """
        if "fcp_p" not in self._hours.columns:
            self._hours["fcp_p"] = self.demand / self.capacity
        return self._hours["fcp_p"]

    @property
    def cons_nom_calef(self) -> "Series[float]":
        """Nominal consumption for heating in kW.
        [Hourly simulation cell FG93]
        """
        return self.nominal_capacity / self._attr("efficiency_cop")

    @property
    def con_cal_t(self) -> "Series[float]":
        """Temperature-dependent consumption. ??????
        [Hourly simulation column FJ]
        """
        if "con_cal_t" not in self._hours.columns:
            self._hours["con_cal_t"] = (
                1.201222828
                - 0.040063338 * self.climate.wet_bulb_temp
                + 0.0010877 * self.climate.wet_bulb_temp**2
            )
        return self._hours["con_cal_t"]

    @property
    def con_cal_fcp(self) -> "Series[float]":
        """Part-load factor for heating system. ??????
        [Hourly simulation column FK]
        """
        if "con_cal_fcp" not in self._hours.columns:
            self._hours["con_cal_fcp"] = (
                0.08565215
                + 0.93881371 * self.fcp_cal
                - 0.1834361 * self.fcp_cal**2
                + 0.15897022 * self.fcp_cal**3
            )
        return self._hours["con_cal_fcp"]

    @property
    def fcp_cal(self) -> "Series[float]":
        """Part-load factor for heating system. ??????
        [Hourly simulation column FL]
        """
        if "fcp_cal" not in self._hours.columns:
            self._hours["fcp_cal"] = self.demand / self.nominal_capacity
        return self._hours["fcp_cal"]

    @property
    def energy_use(self) -> "Series[float]":
        """Heating system energy use in kWh for each hour of the year for each ENERGY_SOURCES.
        [Hourly outputs column P, disaggregated]
        """
        if self._energy_use[self._attr("energy_source")].hasnans:
            self._energy_use = self._energy_use.fillna(0.0).infer_objects(copy=False)
            # Radiators [Hourly simulation column EU]
            if self._attr("type") == HEATING_SYSTEM_TYPES.Electric_heating_radiators:
                # Excel divides demand by nominal capacity in Hourly simulation column EV
                #  Then multiplies again in column EU. These cancel out.
                energy_use = self.demand
            elif self._attr("type") == HEATING_SYSTEM_TYPES.Heat_pump:
                # Heat pump [Hourly simulation column FI]
                energy_use = (self.demand > 0.0) * (
                    self.cons_nom_calef * self.con_cal_t * self.con_cal_fcp
                )
            else:
                # Boilers [Hourly simulation column FB]
                energy_use = (self.Rend != 0.0) * (self.demand / self.Rend)
            self._energy_use[self._attr("energy_source")] = energy_use
        return self._energy_use


class HeatingSimulation(EnergyUseSimulation):
    """A class to simulate heating energy consumption based on building specifications.

    Aggregates the energy use of all heating systems in the building.
    """

    heating_simulations: List[HeatingSystemSimulation]

    def __init__(
        self,
        spec: OpenBESSpecification,
        geometry: BuildingGeometry = None,
        occupancy: OccupationSimulation = None,
        lighting: LightingSimulation = None,
        ventilation: VentilationSimulation = None,
        climate: ClimateSimulation = None,
    ):
        super().__init__(spec)
        try:
            geometry = geometry or BuildingGeometry(self.spec)
            occupancy = occupancy or OccupationSimulation(self.spec, geometry=geometry)
            lighting = lighting or LightingSimulation(self.spec, occupancy=occupancy)
            ventilation = ventilation or VentilationSimulation(
                self.spec, occupancy=occupancy, geometry=geometry
            )
            self.climate_simulation = climate or ClimateSimulation(
                spec,
                geometry=geometry,
                occupancy=occupancy,
                lighting=lighting,
                ventilation=ventilation,
            )
        except SimulationError as exc:
            raise EnergyUseSimulationInitError(
                f"Failed to initialize HeatingSimulation due to error in dependent simulation: {exc}"
            ) from exc
        self.heating_simulations = []
        while True:
            system_number = len(self.heating_simulations) + 1
            try:
                self.heating_simulations.append(
                    HeatingSystemSimulation(
                        spec=spec,
                        system_number=system_number,
                        climate=self.climate_simulation,
                    )
                )
            except EnergyUseSimulationInitError:
                break

    @property
    def energy_use(self) -> "Series[float]":
        """Heating energy use in kWh for each hour of the year for each ENERGY_SOURCES.
        [Hourly outputs column P], disaggregated
        """
        if len(self.heating_simulations) == 0:
            return self._energy_use.fillna(0.0)
        return sum([x.energy_use for x in self.heating_simulations])

    @property
    def report(self) -> HeatingSimulationOutput:
        try:
            systems = []
            precision = output_precision(self.spec)
            area = self.climate_simulation.geometry.conditioned_floor_area
            demand = self.climate_simulation.heating_demand * area / 1000
            quantile = demand.quantile(0.996)
            for sys in self.heating_simulations:
                energy_use = sys.energy_use.sum().sum()
                capacity = sys.nominal_capacity
                systems.append(
                    ThermalSystemResult(
                        conditioned_area=sys.area,
                        energy_demand=energy_use / sys.area if sys.area else None,
                        system_usage=sys.demand.sum() / area if area else None,
                        peak_load=find_hour_peak(demand, max, precision),
                        peak_capacity=capacity,
                        peak_ratio=capacity / quantile if quantile else None,
                    )
                )
            return HeatingSimulationOutput(heating_systems=systems)
        except Exception as exc:
            raise HeatingSimulationError("Failed to generate heating report") from exc



import json
from copy import deepcopy
from importlib.resources import files
from pathlib import Path

from .schemas import OpenBESSpecificationV2
from .schemas.conversion import json_to_toml
from .types import OpenBESSpecification

json_path = Path(str(files("openbes.example_data") / "holywell_house.json"))
with open(json_path) as json_data:
    json_content = json.load(json_data)
toml_content = json_to_toml(json_content)

_HOLYWELL_HOUSE_SPEC = OpenBESSpecification.from_toml(toml_content)
_HOLYWELL_HOUSE_SPEC_V2 = OpenBESSpecificationV2(**json_content)


def get_holywell_house_spec() -> OpenBESSpecification:
    """Return a deep copy of the Holywell House v1 specification."""
    return deepcopy(_HOLYWELL_HOUSE_SPEC)


def get_holywell_house_spec_v2() -> OpenBESSpecificationV2:
    """Return a deep copy of the Holywell House v2 specification."""
    return deepcopy(_HOLYWELL_HOUSE_SPEC_V2)


# Backwards-compatible module constants.
HOLYWELL_HOUSE_SPEC = get_holywell_house_spec()
HOLYWELL_HOUSE_SPEC_V2 = get_holywell_house_spec_v2()



from .. import logging
from pandas import DataFrame

from .base import EnergyUseSimulation, SimulationError
from .occupancy import OccupationSimulation
from ..schemas import HotWaterSimulationOutput
from ..types import OpenBESSpecification

logger = logging.getLogger(__name__)


class HotWaterSimulationError(SimulationError):
    """Raised when hot water report generation fails."""


class HotWaterSimulation(EnergyUseSimulation):
    """Simulate hot water energy use based on building specifications and occupancy patterns."""
    _required_inputs = ("water_system_energy_source",)
    _required_inputs_error_cls = HotWaterSimulationError

    def __init__(
        self, spec: OpenBESSpecification, occupancy: OccupationSimulation = None
    ):
        super().__init__(spec=spec)
        try:
            self.occupancy = occupancy or OccupationSimulation(spec=spec)
        except SimulationError as exc:
            raise HotWaterSimulationError(
                f"Failed to initialize HotWaterSimulation due to error in dependent simulation: {exc}"
            ) from exc

    def get_daily_hot_water_nominal(self) -> float:
        """Calculate nominal (pre-efficiency scaling) daily hot water energy consumption.
        Args:
            spec (OpenBESSpecification): The building specifications spec data class.
        Returns:
            float: Nominal daily hot water energy consumption in kWh.
        """
        demand = self.spec.water_demand  # l/day
        specific_heat_capacity_water = 4.18  # J/g°C
        output_temp = self.spec.water_reference_temperature  # °C
        input_temp = self.spec.water_supply_temperature  # °C
        per_hour = 1 / 3_600  # Convert seconds to hours

        if demand is None or output_temp is None or input_temp is None:
            logger.warning(
                "Insufficient data to calculate hot water energy consumption."
            )
            return 0.0

        temperature_rise = output_temp - input_temp  # °C (from cold to hot water)

        return specific_heat_capacity_water * temperature_rise * demand * per_hour

    def get_daily_hot_water(self) -> float:
        """
        Adjust nominal hot water energy consumption by heater efficiency.
        Args:
            spec (OpenBESSpecification): The building specifications spec data class.
        Returns:
            float: Daily hot water energy consumption in kWh.
        """
        if self.spec.water_system_efficiency_cop is None:
            logger.warning(
                "Insufficient data to calculate hot water energy consumption."
            )
            return 0.0

        return (
            self.get_daily_hot_water_nominal() * self.spec.water_system_efficiency_cop
        )

    def get_hot_water_per_month(self) -> DataFrame:
        """Return the amount of energy used heating water for each month of the year.
        Args:
            spec (OpenBESSpecification): The building specifications spec data class.
        Returns:
            DataFrame: Hot water energy consumption in kWh for each month.
        """
        kWh_per_day = self.get_daily_hot_water()
        result = self.occupancy.occupied_days_per_month * kWh_per_day
        result.name = "kWh"
        return result

    @property
    def energy_use(self) -> DataFrame:
        """Hot water energy use in kWh for each hour of the year."""
        if self._energy_use[self.spec.water_system_energy_source].hasnans:
            daily_kWh = self.get_daily_hot_water()
            hourly_kWh = (
                daily_kWh * self.occupancy.occupied_days_per_month.sum()
            ) / self.occupancy.is_occupied.sum()
            self._energy_use[self.spec.water_system_energy_source] = [
                hourly_kWh
            ] * self.occupancy.is_occupied
        return self._energy_use

    @property
    def report(self) -> HotWaterSimulationOutput:
        try:
            area = self.occupancy.geometry.conditioned_floor_area
            return HotWaterSimulationOutput(
                hot_water_demand=self.energy_use.sum().sum() / area if area else None
            )
        except Exception as exc:
            raise HotWaterSimulationError(
                "Failed to generate hot water report"
            ) from exc
